#!/usr/bin/env python

import sip  # For standalone execution
sip.setapi('QString', 2)
sip.setapi('QDate', 2)
sip.setapi('QDateTime', 2)
sip.setapi('QTime', 2)
sip.setapi('QVariant', 2)

from spykeutils.plugin import analysis_plugin, gui_data
from spykeutils import SpykeException

import uuid
import copy
import pickle
from datetime import datetime

import celery
from PyQt4.QtCore import (Qt, QAbstractTableModel, QTimer, pyqtSignal,
                          QModelIndex, QThread)
from PyQt4 import QtGui

from spyderlib.widgets.dicteditor import DictEditor


class TaskModel(QAbstractTableModel):
    """ This model represents the state of tasks which are run remotely
    using celery. Can update the information automatically in regular intervals.
    """
    col_titles = ('Name', 'Started', 'Worker', 'State', 'Progress',
                  'Progress title / Result', 'Task ID', 'Traceback')
    TASK_ID = 6  # Task id column
    TASK_STATE = 3  # Task state column
    DONE_STATES = ('SUCCESS', 'FAILURE', 'REVOKED')

    class StatusWorker(QThread):
        """ Queries state and result from a task asynchronously. """
        update_entry = pyqtSignal(int, str, object, str)

        def __init__(self, i, ar, tid):
            QThread.__init__(self)
            self.index = i
            self.ar = ar
            self.tid = tid

        def run(self):
            i = self.index
            try:
                state = self.ar.state

                if state == 'FAILURE':
                    tr = self.ar.traceback
                    if tr:
                        result = tr
                    else:
                        result = self.ar.result
                else:
                    result = self.ar.result

                self.update_entry.emit(i, state, result, self.tid)
            except TypeError:
                pass

    def __init__(self, celery, interval=1000, parent=None):
        """ Create a Celery task mode.

        :param celery:
            The `celery.Celery` instance to use.
        :param int interval:
            Interval (in milliseconds) between task status updates.
        """
        QAbstractTableModel.__init__(self, parent)

        self.app = celery
        self._tasks = []  # (task_id, worker, name, time, AsyncResult) tuples
        self._table = []  # (state, result) tuples
        self.workers = []  # List of worker threads for updates
        self._regular_update = False
        self.interval = interval

    def rowCount(self, parent=None, *args, **kwargs):
        return len(self._tasks)

    def columnCount(self, parent=None, *args, **kwargs):
        return len(TaskModel.col_titles)

    def data(self, index, role=None):
        r = index.row()
        c = index.column()
        return self.data_rc(r, c, role)

    def data_rc(self, r, c, role):
        state = self._table[r][0]
        tid, worker, name, time, ar = self._tasks[r]

        if role == Qt.UserRole:  # Custom sort role
            if c == 0:
                return name
            if c == 1:
                t = self._tasks[r][3]
                return (t - datetime(1970, 1, 1)).total_seconds()
            if c == 2:
                return worker
            if c == 3:
                return state
            if c == 4:
                if state == 'PROGRESS':
                    return self._table[r][1]
                if state == 'SUCCESS':
                    return 101
                return -1
            if c == 5:
                if state == 'PROGRESS':
                    try:
                        return self._table[r][1]['title']
                    except TypeError:  # Race condition occured when updating
                        return ''
                else:
                    return str(self._table[r][1])
            elif c == 6:
                return tid

        if role == Qt.DisplayRole or role == Qt.EditRole:
            if c == 0:
                return name
            if c == 1:
                return time.strftime('%a, %d %b %Y %H:%M:%S').decode('UTF-8')
            if c == 2:
                return worker

            state = self._table[r][0]
            if c == 3:
                return state
            if c == 4:
                if state == 'PROGRESS':
                    return self._table[r][1]
                if state == 'SUCCESS':
                    return {'status': 'Done', 'current': 100}
            elif c == 5:
                if state == 'PROGRESS':
                    if self._table[r][1] is None:
                        return ''
                    return self._table[r][1]['title']
                else:
                    if type(self._table[r][1]) in (str, unicode):
                        return self._table[r][1].split('\n')[-2]
                    return str(self._table[r][1])
            elif c == 6:
                return tid.split()[0]
            elif c == 7:
                if state == 'PROGRESS':
                    return self._table[r][1]['title']
                else:
                    return self._table[r][1]

        return None

    def headerData(self, section, orientation, role=None):
        if role == Qt.DisplayRole:
            if orientation == Qt.Horizontal:
                return TaskModel.col_titles[section]
        return None

    def update_task_list(self):
        """ Updates the list of tasks to be monitored. Because broadcast
        operation take a long time, this function may run for a second or more.
        """
        for w in self.workers:  # Stop all running update threads
            w.quit()

        self.beginResetModel()
        self._tasks = []
        self._table = []

        query_result = self.app.control.broadcast('get_tasks', reply=True)
        if not query_result:
            raise SpykeException('Could not contact Server!')
        reserved = self.app.control.inspect().reserved()
        other = query_result[0]

        for all_task_ids in (reserved, other):
            for worker, task_ids in all_task_ids.iteritems():
                for tid in task_ids:
                    try:
                        if type(tid) == dict:
                            tid = tid['id']

                        id_parts = tid.split()
                        task_uid = uuid.UUID(hex=id_parts[0])
                        # Get creation time from UUID
                        time = datetime.fromtimestamp(
                            (task_uid.time - 0x01b21dd213814000L) * 100 / 1e9)
                        name = ' '.join(id_parts[1:])
                    except (ValueError, IndexError):
                        continue

                    r = self.app.AsyncResult(tid)
                    if (tid, worker, name, time, r) not in self._tasks:
                        self._tasks.append((tid, worker, name, time, r))

        self._table = [[None] * 2 for _ in xrange(len(self._tasks))]
        self.endResetModel()

        self._update_values()

    def _update_entry(self, i, state, result, tid):
        """ Update data table when worker thread fetched a result. """
        if len(self._tasks) <= i:
            return
        if self._tasks[i][0] != tid:
            return

        self._table[i][0] = state
        self._table[i][1] = result

        # Invalidate data in table
        top_left = self.createIndex(i, 3)
        bottom_right = self.createIndex(i, 5)
        self.dataChanged.emit(top_left, bottom_right)
        top_left = self.createIndex(i, 7)
        bottom_right = self.createIndex(i, 7)
        self.dataChanged.emit(top_left, bottom_right)

    def _remove_thread(self):
        """ Called when a thread is finished to clean up. """
        self.workers.remove(self.sender())

        if not self.workers and self._regular_update:
            QTimer.singleShot(self.interval, self._update_values)

    def _update_values(self):
        """ Start worker thread to fetch updates on running tasks """
        for i, (tid, _, _, _, ar) in enumerate(self._tasks):
            # Do not check tasks with final states
            if self._table[i][0] not in TaskModel.DONE_STATES:
                w = self.StatusWorker(i, ar, tid)
                w.update_entry.connect(self._update_entry)
                w.finished.connect(self._remove_thread)
                w.start()
                self.workers.append(w)

    def remove_task(self, task_id):
        """ Remove a task from the list of visible tasks. """
        idx = None
        for i, e in enumerate(self._tasks):
            if e[0] == task_id:
                idx = i
                break
        if idx is None:
            return

        self.beginRemoveRows(QModelIndex(), idx, idx)
        self._tasks.pop(idx)
        self._table.pop(idx)
        self.endRemoveRows()

    def start_regular_updates(self):
        """ Starts the periodic status updates. """
        self._regular_update = True
        self._update_values()

    def stop_regular_updates(self):
        """ Stops the periodic status updates. """
        self._regular_update = False


class TaskTableView(QtGui.QTableView):
    """ Used to catch showEvent and hideEvent in order to pause updating task
    states when the table is not visible.
    """
    view_visible = pyqtSignal()
    view_hidden = pyqtSignal()

    def __init__(self, parent=None):
        QtGui.QTableView.__init__(self, parent)

    def showEvent(self, e):
        self.view_visible.emit()

    def hideEvent(self, e):
        self.view_hidden.emit()

    def sizeHintForColumn(self, c):
        if c == 0:
            return 120
        elif c == 1:
            return 180
        elif c == 4:
            return 250
        return -1


class TaskDetailsDialog(QtGui.QDialog):
    """ Dialog to show details of a single task.
    """

    class ProgressItemDelegate(QtGui.QAbstractItemDelegate):
        """ Sets the correct progress and text for the progress bar.
        """
        def __init__(self, parent=None):
            QtGui.QAbstractItemDelegate.__init__(self, parent)

        def setEditorData(self, editor, index):
            content = index.data()
            if index.column() not in (4, 7):
                editor.setText(content)
                return

            if index.column() == 4:
                editor.setVisible(True)
                if content is None:
                    status = index.sibling(index.row(), 3).data()
                    if status in ('FAILURE', 'REVOKED'):
                        editor.setRange(0, 1)
                        editor.setValue(1)
                        editor.setFormat('')
                    else:
                        editor.setRange(0, 0)
                    return

                try:
                    progress = int(content['current'])
                    content = content['status'] + ': ' + str(progress) + '%'
                except ValueError:
                    progress = 100

                editor.setRange(0, 100)
                editor.setValue(progress)
                editor.setFormat(content)
            else:  # 7
                if type(content) not in (str, unicode):
                    editor.setText(str(content))
                else:
                    start = content.find('  File "Spyke Plugin"')
                    if start == -1:
                        editor.setText(content)
                    else:
                        s = content.split('\n')[0] + '\n' + content[start:]
                        editor.setText(s)

    def __init__(self, model, index, app, parent=None):
        QtGui.QDialog.__init__(self, parent)

        self.model = model
        self.app = app
        self.task_id = index.sibling(
            index.row(), TaskModel.TASK_ID).data(Qt.UserRole)
        self.create_gui(index)

    def create_gui(self, index):
        self.setWindowTitle('Task Details')
        self.layout = QtGui.QVBoxLayout()
        self.form_layout = QtGui.QFormLayout()
        self.layout.addLayout(self.form_layout)
        self.setLayout(self.layout)

        name = QtGui.QLineEdit()
        name.setReadOnly(True)
        self.form_layout.addRow(QtGui.QLabel('Name'), name)

        id_field = QtGui.QLineEdit()
        id_field.setReadOnly(True)
        id_field.setMinimumWidth(300)
        self.form_layout.addRow(QtGui.QLabel('ID'), id_field)

        start = QtGui.QLineEdit()
        start.setReadOnly(True)
        self.form_layout.addRow(QtGui.QLabel('Start'), start)

        worker = QtGui.QLineEdit()
        worker.setReadOnly(True)
        self.form_layout.addRow(QtGui.QLabel('Worker'), worker)

        self.status = QtGui.QLineEdit()
        self.status.setReadOnly(True)
        self.form_layout.addRow(QtGui.QLabel('Status'), self.status)

        progress = QtGui.QProgressBar()
        progress.setRange(0, 0)
        self.form_layout.addRow(QtGui.QLabel('Progress'), progress)

        font = QtGui.QFont('Monospace')
        font.setStyleHint(font.TypeWriter, font.PreferDefault)
        
        result = QtGui.QTextEdit()
        result.setReadOnly(True)
        result.setMinimumHeight(200)
        result.setLineWrapMode(QtGui.QTextEdit.NoWrap)
        result.setFont(font)
        self.layout.addWidget(QtGui.QLabel('Progress title / Result'))
        self.layout.addWidget(result)

        self.output = QtGui.QTextEdit()
        self.output.setReadOnly(True)
        self.output.setMinimumHeight(200)
        self.output.setEnabled(False)
        self.output.setLineWrapMode(QtGui.QTextEdit.NoWrap)
        self.output.setFont(font)
        self.layout.addWidget(QtGui.QLabel('Output'))
        self.layout.addWidget(self.output)

        query = QtGui.QPushButton('Query')
        query.pressed.connect(self.query_output)
        self.layout.addWidget(query)

        params_button = QtGui.QPushButton('Query parameters')
        params_button.pressed.connect(self.query_params)
        self.layout.addWidget(params_button)
        
        self.mapper = QtGui.QDataWidgetMapper()
        self.mapper.setModel(self.model)
        self.mapper.addMapping(name, 0)
        self.mapper.addMapping(start, 1)
        self.mapper.addMapping(worker, 2)
        self.mapper.addMapping(self.status, 3)
        self.mapper.addMapping(progress, 4)
        self.mapper.addMapping(result, 7)
        self.mapper.addMapping(id_field, 6)
        self.mapper.setItemDelegate(TaskDetailsDialog.ProgressItemDelegate())
        self.set_source(index)

    def query_output(self):
        output = self.app.control.broadcast(
            'get_task_output', task_id=self.task_id, reply=True)
        text = None
        for d in output:
            for k, v in d.iteritems():
                text = v
        if text:
            self.output.setEnabled(True)
            self.output.setText(text)
        else:
            self.output.setEnabled(False)
            self.output.setText('')

    def query_params(self):
        output = self.app.control.broadcast(
            'get_task_parameters', task_id=self.task_id, reply=True)

        result = None
        for d in output:
            for k, v in d.iteritems():
                result = v

        result['configuration'] = pickle.loads(result['configuration'])

        editor = DictEditor(self)
        editor.setup(result, 'View parameters', readonly=True)
        editor.show()
        editor.raise_()
        editor.activateWindow()

    def set_source(self, index):
        self.mapper.setCurrentIndex(index.row())
        self.output.setEnabled(False)
        self.output.setText('')
        self.task_id = index.sibling(
            index.row(), TaskModel.TASK_ID).data(Qt.UserRole)


class TaskMonitorWidget(QtGui.QWidget):
    """ GUI for monitoring and controlling remotely started plugins.
    """

    class ProgressItemDelegate(QtGui.QStyledItemDelegate):
        """ Shows the progress bar inside the TableView.
        """
        def __init__(self, parent=None):
            QtGui.QStyledItemDelegate.__init__(self, parent)

        def paint(self, painter, option, index):
            content = index.data()
            if index.column() != 4 or content is None:
                QtGui.QStyledItemDelegate.paint(self, painter, option, index)
                return

            try:
                progress = int(content['current'])
                content = content['status'] + ': ' + str(progress) + '%'
            except ValueError:
                progress = 100

            bar_style = QtGui.QStyleOptionProgressBar()
            bar_style.rect = option.rect
            bar_style.minimum = 0
            bar_style.maximum = 100
            bar_style.progress = progress
            bar_style.text = content
            bar_style.textVisible = True

            QtGui.QApplication.style().drawControl(
                QtGui.QStyle.CE_ProgressBar, bar_style, painter)

    def __init__(self, app, parent=None):
        QtGui.QWidget.__init__(self, parent)
        self.app = app
        self.model = None
        self.sort_model = None
        self.table_view = None
        self.details = None
        self.create_gui()

    def create_gui(self):
        self.setWindowTitle('Remote Task Monitor')
        self.resize(900, 500)

        self.model = TaskModel(self.app)
        self.sort_model = QtGui.QSortFilterProxyModel()
        self.sort_model.setSourceModel(self.model)
        self.sort_model.dynamicSortFilter = True
        self.sort_model.setSortRole(Qt.UserRole)
        self.sort_model.dataChanged.connect(self.update_buttons_data_changed)

        self.table_view = TaskTableView()
        self.table_view.setModel(self.sort_model)
        self.table_view.view_visible.connect(self.model.start_regular_updates)
        self.table_view.view_hidden.connect(self.model.stop_regular_updates)
        self.table_view.setSortingEnabled(True)
        self.table_view.hideColumn(2)
        self.table_view.hideColumn(6)
        self.table_view.hideColumn(7)
        self.table_view.resizeColumnsToContents()
        self.table_view.setSelectionMode(TaskTableView.SingleSelection)
        self.table_view.setSelectionBehavior(TaskTableView.SelectRows)
        self.table_view.setHorizontalScrollMode(TaskTableView.ScrollPerPixel)
        self.table_view.verticalHeader().hide()
        self.table_view.setItemDelegate(
            TaskMonitorWidget.ProgressItemDelegate())
        self.table_view.selectionModel().selectionChanged.connect(
            self.selection_changed)
        self.table_view.activated.connect(self.show_details)

        button_layout = QtGui.QHBoxLayout()

        refresh_button = QtGui.QPushButton('Refresh all')
        refresh_button.pressed.connect(self.refresh)
        button_layout.addWidget(refresh_button)

        clear_button = QtGui.QPushButton('Clear all finished')
        clear_button.pressed.connect(self.clear_all)
        button_layout.addWidget(clear_button)

        self.details_button = QtGui.QPushButton('Details')
        self.details_button.pressed.connect(self.details_button_pressed)
        self.details_button.setEnabled(False)
        button_layout.addWidget(self.details_button)

        self.clear_single_button = QtGui.QPushButton('Clear')
        self.clear_single_button.pressed.connect(self.clear_single)
        self.clear_single_button.setEnabled(False)
        button_layout.addWidget(self.clear_single_button)

        self.cancel_button = QtGui.QPushButton('Cancel')
        self.cancel_button.pressed.connect(self.kill)
        self.cancel_button.setEnabled(False)
        button_layout.addWidget(self.cancel_button)

        layout = QtGui.QVBoxLayout(self)
        layout.addLayout(button_layout)
        layout.addWidget(self.table_view)
        self.setLayout(layout)

    def current_task_id(self):
        idx = self.table_view.currentIndex()
        if not idx:
            return None
        return idx.sibling(idx.row(), TaskModel.TASK_ID).data(Qt.UserRole)

    def selection_changed(self, new, _):
        if not new.indexes():
            idx = None
        else:
            idx = new.indexes()[0]
        self.update_buttons(idx)

    def update_buttons_data_changed(self):
        self.update_buttons(self.table_view.currentIndex())

    def update_buttons(self, idx):
        state = None
        try:
            state = idx.sibling(idx.row(),
                                TaskModel.TASK_STATE).data(Qt.UserRole)
        except:
            pass

        if state is None:
            self.details_button.setEnabled(False)
            self.clear_single_button.setEnabled(False)
            self.cancel_button.setEnabled(False)
            return

        self.details_button.setEnabled(True)
        if state in ('SUCCESS', 'FAILURE', 'REVOKED'):
            self.clear_single_button.setEnabled(True)
            self.cancel_button.setEnabled(False)
        else:
            self.clear_single_button.setEnabled(False)
            self.cancel_button.setEnabled(True)

    def refresh(self):
        self.model.update_task_list()

    def kill(self):
        tid = self.current_task_id()
        self.app.control.revoke(tid, terminate=True)

    def clear_all(self):
        self.app.control.broadcast('clear_all_finished_tasks')
        self.refresh()

    def clear_single(self):
        tid = self.current_task_id()
        self.app.control.broadcast(
            'clear_finished_task', task_id=tid)
        self.model.remove_task(tid)

    def start(self, interval):
        self.model.interval = int(interval * 1000)
        self.refresh()

        self.table_view.sortByColumn(1, Qt.DescendingOrder)

    def details_button_pressed(self):
        self.show_details(self.table_view.currentIndex())

    def show_details(self, model_index):
        if self.details is None:
            self.details = TaskDetailsDialog(
                self.sort_model, model_index, self.app, self)
            self.details.setWindowModality(Qt.WindowModal)
        else:
            self.details.set_source(model_index)
        self.details.show()


class CeleryMonitorPlugin(analysis_plugin.AnalysisPlugin):
    interval = gui_data.FloatItem('Update interval', unit='s', default=1.0)
    config_file = gui_data.FileOpenItem(
        'Configuration file', default='celeryconfig.py')
    dock = gui_data.BoolItem('Add as dock')

    def __init__(self, *args, **kwargs):
        super(CeleryMonitorPlugin, self).__init__(*args, **kwargs)
        self.widget = None
        self.window = None

    def get_name(self):
        return 'Qt Celery Monitor'

    def monkeypatch(self, app):
        """ Monkeypatch the AMQP backend to allow task backlogs that are longer
        than 1000 items. Not needed when using Redis.
        """
        try:
            if not isinstance(app.backend, celery.backends.amqp.AMQPBackend):
                return
    
            gtm = app.backend.get_task_meta
            # Just increases the default backlog limit value
    
            def _monkey_patch_get_task_meta(task_id, backlog_limit=10000000):
                return gtm(task_id, backlog_limit)
            app.backend.get_task_meta = _monkey_patch_get_task_meta
        except AttributeError:
            # celery.backends has no attribute amqp if AMQPBackend is not used
            pass

    def start(self, current, selections):
        if self.config_file is None:
            raise SpykeException('Please define a congiguration file!')

        current.progress.begin('Creating task monitor...')

        if self.widget is None:
            self.app = celery.Celery('tasks')
            lcl = {}
            execfile(self.config_file, {}, lcl)
            self.app.config_from_object(lcl)
            self.monkeypatch(self.app)

            self.widget = TaskMonitorWidget(self.app)
            self.window = self.widget

            if self.dock:
                from spykeviewer import api

                self.window = QtGui.QDockWidget('Remote tasks')
                self.window.setObjectName('taskDock')
                self.window.setWidget(self.widget)
                api.window.addDockWidget(
                    Qt.RightDockWidgetArea, self.window)
                self.window.setVisible(True)

        self.widget.start(self.interval)

        current.progress.done()
        self.window.show()
        self.window.raise_()

if __name__ == '__main__':
    from spykeutils.progress_indicator import ProgressIndicator

    class C:
        def __init__(self):
            self.progress = ProgressIndicator()

    app = QtGui.QApplication([])
    p = CeleryMonitorPlugin()
    p.start(C(), None)
    app.exec_()
