import sys
import os
import inspect
import json
import pickle
import multiprocessing
import logging
import socket

import celery
from celery.worker.control import Panel
from celery.signals import task_prerun

from spykeutils.plugin.analysis_plugin import AnalysisPlugin
from spykeutils.plugin.data_provider import DataProvider
from spykeutils.plugin import io_plugin
from spykeutils import progress_indicator

import spykeutils.plugin.data_provider_stored


class CeleryProgress(progress_indicator.ProgressIndicator):
    """ Implementation of a progress indicator that sets the state of the
    corresponding celery task. The state name is 'progress', its meta
    dictionary contains three items: 'current' is the current progress (a value
    between 0 and 100), 'title' is the name of the current operation (as
    specified in the `begin` call and 'status' is the name of the current
    execution status.
    """
    def __init__(self, task):
        super(CeleryProgress, self).__init__()
        self.task = task
        self.current = 0
        self.total = 0
        self.title = ''
        self.status = ''

    def set_ticks(self, ticks):
        self.current = 0
        self.total = ticks

    def begin(self, title=''):
        self.current = 0
        self.total = 0
        self.title = title

    def set_status(self, new_status):
        self.status = new_status

    def done(self):
        pass

    def step(self, num_steps=1):
        self.current += num_steps
        self._update()

    def _update(self):
        c = 0
        if self.total > 0:
            c = float(self.current) / self.total * 100.0
        self.task.update_state(
            state='PROGRESS',
            meta={'current': c, 'title': self.title, 'status': self.status})


def main(args, task):
    """ Start a Spyke Viewer plugin.

    :param dict args:
        Dictionary containing plugin code and parameters.

    :param task:
        The `celery.Task` object where the plugin is started. Used to construct
        a `CeleryProgress` object.
    """
    exc_globals = {}
    code = compile(args['file'], 'Spyke Plugin', 'exec')
    exec(code, exc_globals)

    # Load plugin
    plugin = None
    for cl in exc_globals.values():
        if not inspect.isclass(cl):
            continue

        if not issubclass(cl, AnalysisPlugin):
            continue

        if not cl.__name__ == args['name']:
            continue

        plugin = cl()
        break

    if not plugin:
        sys.stderr.write('Could not find plugin class, aborting...\n')
        sys.exit(1)

    # Load configuration
    if args['config']:
        plugin.set_parameters(pickle.loads(args['config']))

    # Load IO plugins
    if 'io' in args:
        for io in args['io']:
            io_plugin.load_from_string(io)

    # Load selection
    sel_string = args['selection']
    sels = json.loads(sel_string)

    selections = []
    for s in sels:
        selections.append(DataProvider.from_data(s))

    selections[0].progress = CeleryProgress(task)
    plugin.start(selections[0], selections[1:])

    for s in selections:
        if hasattr(s, 'db'):
            s.db.close()


# Celery
# All running and completed tasks are stored in task_params (with task id as
# key and parameter dictionary as value). Completed tasks include failed and
# aborted tasks.
manager = multiprocessing.Manager()
task_params = manager.dict()
task_ids_lock = manager.Lock()

app = celery.Celery('spyke')
lcl = {}
execfile('celeryconfig.py', {}, lcl)
app.config_from_object(lcl)

# Monkeypatch amqp backend to support removing task result queues and
# dealing with queues containing very many messages
try:
    if isinstance(app.backend, celery.backends.amqp.AMQPBackend):
        from kombu import Connection, Queue
        _kombu_connection = Connection(lcl['BROKER_URL'])

        def _monkey_patch_forget(task_id):
            q = Queue(name=task_id,
                      exchange=app.backend.exchange,
                      routing_key=task_id,
                      durable=app.backend.persistent,
                      queue_arguments=app.backend.queue_arguments,
                      auto_delete=app.backend.auto_delete)(_kombu_connection)
            q.delete()
        app.backend._forget = _monkey_patch_forget

        _gtm = app.backend.get_task_meta

        # Just increases the default backlog_limit value
        def _monkey_patch_get_task_meta(task_id, backlog_limit=10000000):
            return _gtm(task_id, backlog_limit)
        app.backend.get_task_meta = _monkey_patch_get_task_meta
except AttributeError:
    pass  # celery.backends has no attribute amqp if AMQPBackend is not used


@Panel.register
def get_tasks(_):
    """ Return a list of task ids of all known running and completed tasks.
    """
    task_ids_lock.acquire()
    retval = task_params.keys()
    task_ids_lock.release()
    return retval


# Completely forget a task
def _remove_task(task_id):  # Call only when task_ids_lock is acquired!
    try:
        del task_params[task_id]
    except:
        pass

    try:
        app.backend.forget(task_id)
    finally:
        try:
            os.remove(task_id)
        except:
            pass


@Panel.register
def clear_finished_task(_, task_id):
    """ Removes a completed task. Does nothing if the task is not completed.
    """
    task_ids_lock.acquire()
    try:
        r = app.AsyncResult(task_id)
        if r.ready() or r.state == 'PENDING':
            _remove_task(task_id)
    # No state for revoked, TypeError with failed, remove anyway
    except (socket.error, TypeError):
        _remove_task(t)
    finally:
        task_ids_lock.release()


@Panel.register
def clear_all_finished_tasks(_):
    """ Removes all completed tasks. Does not touch tasks that are not
    completed.
    """
    task_ids_lock.acquire()
    try:
        for t in task_params.keys()[:]:
            try:
                r = app.AsyncResult(t)
                if r.ready() or r.state == 'PENDING':
                    _remove_task(t)
            # No state for revoked, TypeError with failed, remove anyway
            except (socket.error, TypeError):
                _remove_task(t)
    finally:
        task_ids_lock.release()


@Panel.register
def get_task_output(_, task_id):
    """ Returns all stdout/stderr output of a task as a string.
    """
    try:
        f = open(task_id, 'r')
        t = f.read()
        f.close()
        return t
    except:
        return None


@Panel.register
def get_task_parameters(_, task_id):
    """ Returns a dictionary with the parameters of a task. For plugins, it
    contains two items: 'selections' with the selected data, including
    the current selection, and 'configuration' with the guidata configuration
    of the plugin.
    """
    return task_params[task_id]


# Add task id and its parameters (if the task is a plugin) when a task is
# about to run.
@task_prerun.connect
def add_task_to_list(task_id, *args, **kwargs):
    task_ids_lock.acquire()
    if task_id not in task_params:
        try:
            selections = json.loads(kwargs['args'][0]['selection'])
            config = kwargs['args'][0]['config']
            task_params[task_id] = {'selections': selections,
                                    'configuration': config}
        except (KeyError, IndexError):
            task_params[task_id] = {'selections': [], 'configuration': {}}
    task_ids_lock.release()


class SpykePlugin(app.Task):
    """ A Celery task to run a Spyke Viewer plugin.
    """
    name = 'SpykePlugin'

    def run(self, args):
        log = logging.getLogger()
        handler = logging.FileHandler(self.request.id, 'w')
        log.setLevel(logging.WARNING)
        log.addHandler(handler)

        if os.getcwd() not in sys.path:
            sys.path.insert(0, os.getcwd())

        try:
            main(args, self)
        finally:
            log.removeHandler(handler)
            handler.close()
